import json
import time

from aliyun.log import LogClient, LogItem, PutLogsRequest

from cobweb.base import Queue, logger
from cobweb import setting


class LoghubDot:

    def __init__(self):
        self.client = LogClient(**setting.LOGHUB_CONFIG)
        self.queue = Queue()

    def build(self, topic, **kwargs):

        temp = {}
        log_item = LogItem()
        for key, value in kwargs.items():
            if not isinstance(value, str):
                temp[key] = json.dumps(value, ensure_ascii=False)
            else:
                temp[key] = value
        contents = sorted(temp.items())
        log_item.set_contents(contents)
        self.queue.push((topic, log_item), direct_insertion=True)

    def build_run(self):
        while True:
            start_time = int(time.time())
            while True:
                cost_time = int(time.time()) - start_time
                if self.queue.length >= 1000 or cost_time > 10:
                    break
                time.sleep(0.5)
            try:
                log_item_info = {}
                for _ in range(1000):
                    its = self.queue.pop()
                    if not its:
                        break
                    topic, item = its
                    log_item_info.setdefault(topic, []).append(item)
                for topic, log_items in log_item_info.items():
                    request = PutLogsRequest(
                        project="databee-download-log",
                        logstore="log",
                        topic=topic,
                        logitems=log_items,
                        compress=True
                    )
                    self.client.put_logs(request=request)
            except Exception as e:
                logger.info(str(e))



import json
import time
import hashlib


class SeedParams:

    def __init__(self, retry, priority, seed_version, seed_status=None, proxy_type=None, proxy=None):
        self.retry = retry or 0
        self.priority = priority or 300
        self.seed_version = seed_version or int(time.time())
        self.seed_status = seed_status
        self.proxy_type = proxy_type
        self.proxy = proxy


class Seed:

    __SEED_PARAMS__ = [
        "retry",
        "priority",
        "seed_version",
        "seed_status",
        "proxy_type",
        "proxy"
    ]

    def __init__(
            self,
            seed,
            sid=None,
            retry=None,
            priority=None,
            seed_version=None,
            seed_status=None,
            proxy_type=None,
            proxy=None,
            **kwargs
    ):
        if any(isinstance(seed, t) for t in (str, bytes)):
            try:
                item = json.loads(seed)
                self._init_seed(item)
            except json.JSONDecodeError:
                self.__setattr__("url", seed)
        elif isinstance(seed, dict):
            self._init_seed(seed)
        else:
            raise TypeError(Exception(
                f"seed type error, "
                f"must be str or dict! "
                f"seed: {seed}"
            ))

        seed_params = {
            "retry": retry,
            "priority": priority,
            "seed_version": seed_version,
            "seed_status": seed_status,
            "proxy_type": proxy_type,
            "proxy": proxy
        }

        if kwargs:
            self._init_seed(kwargs)
            seed_params.update({
                k:v for k, v in kwargs.items()
                if k in self.__SEED_PARAMS__
            })
        if sid or not getattr(self, "sid", None):
            self._init_id(sid)
        self.params = SeedParams(**seed_params)

    def __getattr__(self, name):
        return None

    def __setitem__(self, key, value):
        setattr(self, key, value)

    def __getitem__(self, item):
        return getattr(self, item)

    def __str__(self):
        return json.dumps(self.__dict__, ensure_ascii=False)

    def __repr__(self):
        chars = [f"{k}={v}" for k, v in self.__dict__.items()]
        return f'{self.__class__.__name__}({", ".join(chars)})'

    def _init_seed(self, seed_info:dict):
        for k, v in seed_info.items():
            if k not in self.__SEED_PARAMS__:
                self.__setattr__(k, v)

    def _init_id(self, sid):
        if not sid:
            sid = hashlib.md5(self.to_string.encode()).hexdigest()
        self.__setattr__("sid", sid)

    @property
    def to_dict(self) -> dict:
        seed = self.__dict__.copy()
        if seed.get("params"):
            del seed["params"]
        return seed

    @property
    def to_string(self) -> str:
        return json.dumps(
            self.to_dict,
            ensure_ascii=False,
            separators=(",", ":")
        )

    @property
    def get_all(self):
        return json.dumps(
            self.__dict__,
            ensure_ascii=False,
            separators=(",", ":")
        )





from cobweb import setting
from requests import Response
from oss2 import Auth, Bucket, models, PartIterator
from cobweb.exceptions import oss_db_exception
from cobweb.base.decorators import decorator_oss_db


class OssUtil:

    def __init__(
        self,
        bucket=None,
        endpoint=None,
        access_key=None,
        secret_key=None,
        chunk_size=None,
        min_upload_size=None,
        **kwargs
    ):
        self.bucket = bucket or setting.OSS_BUCKET
        self.endpoint = endpoint or setting.OSS_ENDPOINT
        self.chunk_size = int(chunk_size or setting.OSS_CHUNK_SIZE)
        self.min_upload_size = int(min_upload_size or setting.OSS_MIN_UPLOAD_SIZE)

        self.failed_count = 0
        self._kw = kwargs

        self._auth = Auth(
            access_key_id=access_key or setting.OSS_ACCESS_KEY,
            access_key_secret=secret_key or setting.OSS_SECRET_KEY
        )
        self._client = Bucket(
            auth=self._auth,
            endpoint=self.endpoint,
            bucket_name=self.bucket,
            **self._kw
        )

    def failed(self):
        self.failed_count += 1
        if self.failed_count >= 5:
            self._client = Bucket(
                auth=self._auth,
                endpoint=self.endpoint,
                bucket_name=self.bucket,
                **self._kw
            )

    def exists(self, key: str) -> bool:
        try:
            result = self._client.object_exists(key)
            self.failed_count = 0
            return result
        except Exception as e:
            self.failed()
            raise e

    def head(self, key: str) -> models.HeadObjectResult:
        return self._client.head_object(key)

    @decorator_oss_db(exception=oss_db_exception.OssDBInitPartError)
    def init_part(self, key) -> models.InitMultipartUploadResult:
        """初始化分片上传"""
        return self._client.init_multipart_upload(key)

    @decorator_oss_db(exception=oss_db_exception.OssDBPutObjError)
    def put(self, key, data) -> models.PutObjectResult:
        """文件上传"""
        return self._client.put_object(key, data)

    @decorator_oss_db(exception=oss_db_exception.OssDBPutPartError)
    def put_part(self, key, upload_id, position, data) -> models.PutObjectResult:
        """分片上传"""
        return self._client.upload_part(key, upload_id, position, data)

    def list_part(self, key, upload_id):  # -> List[models.ListPartsResult]:
        """获取分片列表"""
        return [part_info for part_info in PartIterator(self._client, key, upload_id)]

    @decorator_oss_db(exception=oss_db_exception.OssDBMergeError)
    def merge(self, key, upload_id, parts=None) -> models.PutObjectResult:
        """合并分片"""
        headers = None if parts else {"x-oss-complete-all": "yes"}
        return self._client.complete_multipart_upload(key, upload_id, parts, headers=headers)

    @decorator_oss_db(exception=oss_db_exception.OssDBAppendObjError)
    def append(self, key, position, data) -> models.AppendObjectResult:
        """追加上传"""
        return self._client.append_object(key, position, data)

    def iter_data(self, data, chunk_size=None):
        chunk_size = chunk_size or self.chunk_size
        if isinstance(data, Response):
            for part_data in data.iter_content(chunk_size):
                yield part_data
        if isinstance(data, bytes):
            for i in range(0, len(data), chunk_size):
                yield data[i:i + chunk_size]

    def assemble(self, ready_data, data, chunk_size=None):
        upload_data = b""
        ready_data = ready_data + data
        chunk_size = chunk_size or self.chunk_size
        if len(ready_data) >= chunk_size:
            upload_data = ready_data[:chunk_size]
            ready_data = ready_data[chunk_size:]
        return ready_data, upload_data

    def content_length(self, key: str) -> int:
        head = self.head(key)
        return head.content_length




import math
import time

import mmh3
import redis
from cobweb import setting


class BloomFilter:

    def __init__(self, key, redis_config=None, capacity=None, error_rate=None):
        redis_config = redis_config or setting.REDIS_CONFIG
        capacity = capacity or setting.CAPACITY
        error_rate = error_rate or setting.ERROR_RATE
        redis_config['db'] = 3

        self.key = key

        pool = redis.ConnectionPool(**redis_config)
        self._client = redis.Redis(connection_pool=pool)
        self.bit_size = self.get_bit_size(capacity, error_rate)
        self.hash_count = self.get_hash_count(self.bit_size, capacity)
        self._init_bloom_key()

    def add(self, value):
        for seed in range(self.hash_count):
            result = mmh3.hash(value, seed) % self.bit_size
            self._client.setbit(self.key, result, 1)
        return True

    def exists(self, value):
        if not self._client.exists(self.key):
            return False
        for seed in range(self.hash_count):
            result = mmh3.hash(value, seed) % self.bit_size
            if not self._client.getbit(self.key, result):
                return False
        return True

    def _init_bloom_key(self):
        lua_script = """
        redis.call("SETBIT", KEYS[1], ARGV[1], ARGV[2])
        redis.call("EXPIRE", KEYS[1], 604800)
        """
        if self._client.exists(self.key):
            return True
        execute = self._client.register_script(lua_script)
        execute(keys=[self.key], args=[self.bit_size-1, 1])

    @classmethod
    def get_bit_size(cls, n, p):
        return int(-(n * math.log(p)) / (math.log(2) ** 2))

    @classmethod
    def get_hash_count(cls, m, n):
        return int((m / n) * math.log(2))





import os
from typing import Union
from cobweb import setting
from cobweb.utils import OssUtil
from cobweb.crawlers import Crawler
from cobweb.base import Seed, BaseItem, Request, Response
from cobweb.exceptions import OssDBPutPartError, OssDBMergeError


oss_util = OssUtil(is_path_style=bool(int(os.getenv("PRIVATE_LINK", 0))))


class FileCrawlerAir(Crawler):

    @staticmethod
    def download(item: Request) -> Union[Seed, BaseItem, Response, str]:
        seed_dict = item.seed.to_dict
        seed_dict["bucket_name"] = oss_util.bucket
        try:
            seed_dict["oss_path"] = key = item.seed.oss_path or getattr(item, "oss_path")

            if oss_util.exists(key):
                seed_dict["data_size"] = oss_util.head(key).content_length
                yield Response(item.seed, "exists", **seed_dict)

            else:
                seed_dict.setdefault("end", "")
                seed_dict.setdefault("start", 0)

                if seed_dict["end"] or seed_dict["start"]:
                    start, end = seed_dict["start"], seed_dict["end"]
                    item.request_setting["headers"]['Range'] = f'bytes={start}-{end}'

                if not item.seed.identifier:
                    content = b""
                    chunk_size = oss_util.chunk_size
                    min_upload_size = oss_util.min_upload_size
                    seed_dict.setdefault("position", 1)

                    response = item.download()

                    content_type = response.headers.get("content-type", "").split(";")[0]
                    seed_dict["data_size"] = content_length = int(response.headers.get("content-length", 0))

                    if content_type and content_type in setting.FILE_FILTER_CONTENT_TYPE:
                        """过滤响应文件类型"""
                        response.close()
                        seed_dict["filter"] = True
                        seed_dict["msg"] = f"response content type is {content_type}"
                        yield Response(item.seed, response, **seed_dict)

                    elif seed_dict['position'] == 1 and min_upload_size >= content_length > 0:
                        """过小文件标识返回"""
                        response.close()
                        seed_dict["filter"] = True
                        seed_dict["msg"] = "file size is too small"
                        yield Response(item.seed, response, **seed_dict)

                    elif seed_dict['position'] == 1 and chunk_size > content_length > min_upload_size:
                        """小文件直接下载"""
                        for part_data in response.iter_content(chunk_size):
                            content += part_data
                        response.close()
                        oss_util.put(key, content)
                        yield Response(item.seed, response, **seed_dict)

                    else:
                        """中大文件同步分片下载"""
                        seed_dict.setdefault("upload_id", oss_util.init_part(key).upload_id)

                        for part_data in response.iter_content(chunk_size):
                            content += part_data
                            if len(content) >= chunk_size:
                                upload_data = content[:chunk_size]
                                content = content[chunk_size:]
                                oss_util.put_part(key, seed_dict["upload_id"], seed_dict['position'], content)
                                seed_dict['start'] += len(upload_data)
                                seed_dict['position'] += 1

                        response.close()

                        if content:
                            oss_util.put_part(key, seed_dict["upload_id"], seed_dict['position'], content)
                        oss_util.merge(key, seed_dict["upload_id"])
                        seed_dict["data_size"] = oss_util.head(key).content_length
                        yield Response(item.seed, response, **seed_dict)

                elif item.seed.identifier == "merge":
                    oss_util.merge(key, seed_dict["upload_id"])
                    seed_dict["data_size"] = oss_util.head(key).content_length
                    yield Response(item.seed, "merge", **seed_dict)

        except OssDBPutPartError:
            yield Seed(seed_dict)
        except OssDBMergeError:
            yield Seed(seed_dict, identifier="merge")





import time
import inspect
import threading
import importlib
from functools import wraps


from cobweb import setting
from cobweb.base import Seed, Queue, logger
from cobweb.utils.tools import dynamic_load_class


def check_pause(func):
    @wraps(func)
    def wrapper(self, *args, **kwargs):
        while not self._pause.is_set():
            try:
                func(self, *args, **kwargs)
            except Exception as e:
                logger.info(f"{func.__name__}: " + str(e))
            finally:
                time.sleep(0.1)

    return wrapper


class Launcher(threading.Thread):

    SEEDS = []

    __DOING__ = {}

    __CUSTOM_FUNC__ = {
        # "download": None,
        # "request": None,
        # "parse": None,
    }

    __LAUNCHER_QUEUE__ = {
        "new": Queue(),
        "todo": Queue(),
        "done": Queue(),
        "upload": Queue()
    }

    __LAUNCHER_FUNC__ = [
        "_reset",
        "_scheduler",
        "_insert",
        "_refresh",
        "_delete",
    ]

    __WORKER_THREAD__ = {}

    def __init__(self, task, project, custom_setting=None, **kwargs):
        super().__init__()
        self.task = task
        self.project = project

        self._app_time = int(time.time())
        self._stop = threading.Event()  # 结束事件
        self._pause = threading.Event()  # 暂停事件

        _setting = dict()

        if custom_setting:
            if isinstance(custom_setting, dict):
                _setting = custom_setting
            else:
                if isinstance(custom_setting, str):
                    custom_setting = importlib.import_module(custom_setting)
                if not inspect.ismodule(custom_setting):
                    raise Exception
                for k, v in custom_setting.__dict__.items():
                    if not k.startswith("__") and not inspect.ismodule(v):
                        _setting[k] = v

        _setting.update(**kwargs)

        for k, v in _setting.items():
            setattr(setting, k.upper(), v)

        self._Crawler = dynamic_load_class(setting.CRAWLER)
        self._Pipeline = dynamic_load_class(setting.PIPELINE)

        self._before_scheduler_wait_seconds = setting.BEFORE_SCHEDULER_WAIT_SECONDS
        self._scheduler_wait_seconds = setting.SCHEDULER_WAIT_SECONDS
        self._todo_queue_full_wait_seconds = setting.TODO_QUEUE_FULL_WAIT_SECONDS
        self._new_queue_wait_seconds = setting.NEW_QUEUE_WAIT_SECONDS
        self._done_queue_wait_seconds = setting.DONE_QUEUE_WAIT_SECONDS
        self._upload_queue_wait_seconds = setting.UPLOAD_QUEUE_WAIT_SECONDS
        self._seed_reset_seconds = setting.SEED_RESET_SECONDS

        self._todo_queue_size = setting.TODO_QUEUE_SIZE
        self._new_queue_max_size = setting.NEW_QUEUE_MAX_SIZE
        self._done_queue_max_size = setting.DONE_QUEUE_MAX_SIZE
        self._upload_queue_max_size = setting.UPLOAD_QUEUE_MAX_SIZE

        self._spider_max_retries = setting.SPIDER_MAX_RETRIES
        self._spider_thread_num = setting.SPIDER_THREAD_NUM
        self._spider_time_sleep = setting.SPIDER_TIME_SLEEP
        self._spider_max_count = setting.SPIDER_MAX_COUNT
        self._time_window = setting.TIME_WINDOW
        self._speed_control = setting.SPEED_CONTROL

        self._done_model = setting.DONE_MODEL
        self._task_model = setting.TASK_MODEL

        self._filter_field = setting.FILTER_FIELD

    @property
    def request(self):
        """
        自定义request函数
        use case:
            from cobweb.base import Request, BaseItem
            @launcher.request
            def request(seed: Seed) -> Union[Request, BaseItem]:
                ...
                yield Request(seed.url, seed)
        """
        def decorator(func):
            self.__CUSTOM_FUNC__["request"] = func
        return decorator

    @property
    def download(self):
        """
        自定义download函数
        use case:
            from cobweb.base import Request, Response, Seed, BaseItem
            @launcher.download
            def download(item: Request) -> Union[Seed, BaseItem, Response, str]:
                ...
                yield Response(item.seed, response)
        """
        def decorator(func):
            self.__CUSTOM_FUNC__["download"] = func
        return decorator

    @property
    def parse(self):
        """
        自定义parse函数, xxxItem为自定义的存储数据类型
        use case:
            from cobweb.base import Request, Response
            @launcher.parse
            def parse(item: Response) -> BaseItem:
               ...
               yield xxxItem(seed, **kwargs)
        """
        def decorator(func):
            self.__CUSTOM_FUNC__["parse"] = func
        return decorator

    def start_seeds(self):
        seeds = [Seed(seed) for seed in self.SEEDS]
        self.__LAUNCHER_QUEUE__['todo'].push(seeds)
        return seeds

    def _remove_doing_seeds(self, seeds):
        for seed in seeds:
            self.__DOING__.pop(seed, None)
        # logger.info("remove %s seeds from __DOING__" % len(seeds))

    def _get_seed(self) -> Seed:
        return self.__LAUNCHER_QUEUE__["todo"].pop()

    def _set_seed(self, seed, **kwargs):
        self.__LAUNCHER_QUEUE__["todo"].push(seed, **kwargs)

    def _upload_data(self, data, **kwargs):
        self.__LAUNCHER_QUEUE__["upload"].push(data, **kwargs)

    def _add_seed(self, seed, **kwargs):
        self.__LAUNCHER_QUEUE__["new"].push(seed, **kwargs)

    def _delete_seed(self, seed, **kwargs):
        self.__LAUNCHER_QUEUE__["done"].push(seed, **kwargs)

    def _execute(self):
        for func_name in self.__LAUNCHER_FUNC__:
            worker_thread = threading.Thread(name=func_name, target=getattr(self, func_name))
            self.__WORKER_THREAD__[func_name] = worker_thread
            worker_thread.start()
            time.sleep(1)

    def _monitor(self):
        while True:
            if not self.__WORKER_THREAD__:
                time.sleep(15)
                continue
            for func_name, worker_thread in self.__WORKER_THREAD__.items():
                if not worker_thread.is_alive():
                    logger.info(f"{func_name} thread is dead. Restarting...")
                    target = getattr(self, func_name)
                    worker_thread = threading.Thread(name=func_name, target=target)
                    self.__WORKER_THREAD__[func_name] = worker_thread
                    worker_thread.start()
                time.sleep(5)

    def run(self):
        threading.Thread(target=self._execute_heartbeat).start()

        self.start_seeds()

        self._Crawler(
            task=self.task, project=self.project,
            stop=self._stop, pause=self._pause,
            # launcher_queue=self.__LAUNCHER_QUEUE__,
            get_seed=self._get_seed,
            set_seed=self._set_seed,
            add_seed=self._add_seed,
            delete_seed=self._delete_seed,
            upload_data=self._upload_data,
            custom_func=self.__CUSTOM_FUNC__,
            thread_num = self._spider_thread_num,
            max_retries = self._spider_max_retries,
            time_sleep=self._spider_time_sleep
        ).start()

        self._Pipeline(
            stop=self._stop, pause=self._pause,
            upload=self.__LAUNCHER_QUEUE__["upload"],
            done=self.__LAUNCHER_QUEUE__["done"],
            upload_size=self._upload_queue_max_size,
            wait_seconds=self._upload_queue_wait_seconds
        ).start()

        threading.Thread(target=self._monitor).start()

        self._execute()
        self._polling()

    def _execute_heartbeat(self):
        pass

    def _reset(self):
        pass

    def _scheduler(self):
        pass

    def _insert(self):
        pass

    def _refresh(self):
        pass

    def _delete(self):
        pass

    def _polling(self):
        pass

